# -*- coding: iso_8859_1 -*-

import numpy as np
from scipy.optimize import leastsq

#----------------------------------------------------------------------------------------
#def fit_gauss_circular(xy, data):
#def fit_gauss_elliptical(xy, data):
#def fit_moffat_circular(xy, data):
#def fit_moffat_elliptical(xy, data):
#
#--- purpose ---
#these two functions perform a PSF fitting of a stellar object, using a 2D anisotropic Gaussian PSF with arbitrary azimuth orientation (same gaussian PSF fit as the PRISM software is doing)
#
#--- input parameters ---
#data = 2D numpy array containing data to be fitted, with size (2N+1)*(2N+1)
#		the array "data" is typically a subsample of a full image, it must only contain a single object: the one to be fitted, placed approximately at center
#xy = list with two values = [x,y]
#		x and y are the coordinates on the initial full image of the first pixel of the data subimage (the one with x,y = 0,0)
#
#--- output parameters ---
#result = list with 8 elements
#		result[k] gives the k-th PSF fit parameter for the object
#		result[0] gives the max measured value of the object
#		result[1] gives the floor value
#		result[2] gives the height of the 2D gaussian
#		result[3] gives the center position along X axis on the initial image
#		result[4] gives the center position along Y axis on the initial image
#		result[5] gives the (azimuthally) smallest fwhm
#		result[6] gives the fwhm in the perpendicular direction (largest fwhm)
#		result[7] gives the azimuth orientation of the largest fwhm, measured starting from x (vertical axis) clockwise towards y (horizontal axis)
#----------------------------------------------------------------------------------------

def fit_gauss_circular(xy, data):

	#---------------------------------------------------------------------------------
	#find starting values
	floor = np.ma.median(data.flatten())
	height = data.max() - floor

	#print data.max(), height

	data_x = np.sum(data-floor,axis=1)
	data_y = np.sum(data-floor,axis=0)

	mean_x = np.argmax(data_x)
	mean_y = np.argmax(data_y)
	
	fwhm_x = np.float32((data_x > data_x.min()+ (data_x.max()-data_x.min())/2.).sum())
	fwhm_y = np.float32((data_y > data_y.min()+ (data_y.max()-data_y.min())/2.).sum())
	fwhm = 0.5*(fwhm_x+fwhm_y)
	
	p0 = floor, height, mean_x, mean_y, fwhm

	#---------------------------------------------------------------------------------
	#fitting gaussian
	def gauss(floor, height, mean_x, mean_y, fwhm):
		sig = fwhm / (2.*np.sqrt(2.*np.log(2.)))
		C = 0.5/np.square(sig)
		
		return lambda x,y: floor + height*np.exp(-C*((x-mean_x)**2)-C*((y-mean_y)**2))

	def err(p,data):
		return np.ravel(gauss(*p)(*np.indices(data.shape))-data)
	
	p = leastsq(err, p0, args=(data), maxfev=1000)
	p = p[0]
	
	#---------------------------------------------------------------------------------
	#formatting results
	maxi = data.max()
	floor = p[0]
	height = p[1]
	mean_x = p[2] + xy[0]
	mean_y = p[3] + xy[1]
	fwhm = np.abs(p[4])
	
	output = [maxi, floor, height, mean_x, mean_y, fwhm]
	return output

def fit_gauss_elliptical(xy, data):

	#---------------------------------------------------------------------------------
	#find starting values
	floor = np.ma.median(data.flatten())
	height = data.max() - floor

	#print data.max(), height

	data_x = np.sum(data-floor,axis=1)
	data_y = np.sum(data-floor,axis=0)

	mean_x = np.argmax(data_x)
	mean_y = np.argmax(data_y)
	
	fwhm_x = np.float32((data_x > data_x.min()+ (data_x.max()-data_x.min())/2.).sum())
	fwhm_y = np.float32((data_y > data_y.min()+ (data_y.max()-data_y.min())/2.).sum())
	
	angle = 0.
	
	p0 = floor, height, mean_x, mean_y, fwhm_x, fwhm_y, angle

	#---------------------------------------------------------------------------------
	#fitting gaussian
	def gauss(floor, height, mean_x, mean_y, fwhm_x, fwhm_y, angle):
		sig_x = fwhm_x / (2.*np.sqrt(2.*np.log(2.)))
		sig_y = fwhm_y / (2.*np.sqrt(2.*np.log(2.)))	
		
		A = (np.cos(angle)/sig_x)**2. + (np.sin(angle)/sig_y)**2.
		B = (np.sin(angle)/sig_x)**2. + (np.cos(angle)/sig_y)**2.
		C = 2.0*np.sin(angle)*np.cos(angle)*(1./sig_x**2. - 1./sig_y**2.)

		#do not forget factor 0.5 in exp(-0.5*r**2./sig**2.)	
		return lambda x,y: floor + height*np.exp(-0.5*(A*((x-mean_x)**2)+B*((y-mean_y)**2)+C*(x-mean_x)*(y-mean_y)))

	def err(p,data):
		return np.ravel(gauss(*p)(*np.indices(data.shape))-data)
	
	p = leastsq(err, p0, args=(data), maxfev=1000)
	p = p[0]
	
	#---------------------------------------------------------------------------------
	#formatting results
	maxi = data.max()
	floor = p[0]
	height = p[1]
	mean_x = p[2] + xy[0]
	mean_y = p[3] + xy[1]
	fwhm_small = np.min((np.abs(p[4]),np.abs(p[5])))
	fwhm_large = np.max((np.abs(p[4]),np.abs(p[5])))
	angle = p[6]	#gives the direction of the p[4]=sig_x axis, starting from x (vertical) axis, clockwise in direction of y (horizontal) axis
	if p[5]>p[4]:	#then sig_x is the smallest : we want angle to point to sig_y, the largest
		angle = angle + np.pi/2.
	angle = np.arctan(np.tan(angle))
	
	output = [maxi, floor, height, mean_x, mean_y, fwhm_small, fwhm_large, angle]
	return output

def fit_moffat_circular(xy, data):

	#---------------------------------------------------------------------------------
	#find starting values
	floor = np.ma.median(data.flatten())
	height = data.max() - floor

	#print data.max(), height

	data_x = np.sum(data-floor,axis=1)
	data_y = np.sum(data-floor,axis=0)

	mean_x = np.argmax(data_x)
	mean_y = np.argmax(data_y)
	
	fwhm_x = np.float32((data_x > data_x.min()+ (data_x.max()-data_x.min())/2.).sum())
	fwhm_y = np.float32((data_y > data_y.min()+ (data_y.max()-data_y.min())/2.).sum())
	fwhm = 0.5*(fwhm_x+fwhm_y)

	beta = 4
	
	p0 = floor, height, mean_x, mean_y, fwhm, beta

	#---------------------------------------------------------------------------------
	#fitting gaussian
	def moffat(floor, height, mean_x, mean_y, fwhm, beta):
		sig = 0.5*fwhm/np.sqrt(2.**(1./beta)-1.)	
		return lambda x,y: floor + height/((1.+(((x-mean_x)**2+(y-mean_y)**2)/sig**2.))**beta)

	def err(p,data):
		return np.ravel(moffat(*p)(*np.indices(data.shape))-data)
	
	p = leastsq(err, p0, args=(data), maxfev=1000)
	p = p[0]
	
	#---------------------------------------------------------------------------------
	#formatting results
	maxi = data.max()
	floor = p[0]
	height = p[1]
	mean_x = p[2] + xy[0]
	mean_y = p[3] + xy[1]
	fwhm = np.abs(p[4])
	beta = p[5]
	
	output = [maxi, floor, height, mean_x, mean_y, fwhm, beta]
	return output

def fit_moffat_elliptical(xy, data):

	#---------------------------------------------------------------------------------
	#find starting values
	floor = np.ma.median(data.flatten())
	height = data.max() - floor

	#print data.max(), height

	data_x = np.sum(data-floor,axis=1)
	data_y = np.sum(data-floor,axis=0)

	mean_x = np.argmax(data_x)
	mean_y = np.argmax(data_y)
	
	fwhm_x = np.float32((data_x > data_x.min()+ (data_x.max()-data_x.min())/2.).sum())
	fwhm_y = np.float32((data_y > data_y.min()+ (data_y.max()-data_y.min())/2.).sum())

	beta = 4
	angle = 0.
	
	p0 = floor, height, mean_x, mean_y, fwhm_x, fwhm_y, angle, beta

	#---------------------------------------------------------------------------------
	#fitting gaussian
	def moffat(floor, height, mean_x, mean_y, fwhm_x, fwhm_y, angle, beta):
		
		sig_x = 0.5*fwhm_x/np.sqrt(2.**(1./beta)-1.)
		sig_y = 0.5*fwhm_y/np.sqrt(2.**(1./beta)-1.)
	
		A = (np.cos(angle)/sig_x)**2. + (np.sin(angle)/sig_y)**2.
		B = (np.sin(angle)/sig_x)**2. + (np.cos(angle)/sig_y)**2.
		C = 2.0*np.sin(angle)*np.cos(angle)*(1./sig_x**2. - 1./sig_y**2.)
		
		return lambda x,y: floor + height/((1.+ A*((x-mean_x)**2) + B*((y-mean_y)**2) + C*(x-mean_x)*(y-mean_y))**beta)

	def err(p,data):
		return np.ravel(moffat(*p)(*np.indices(data.shape))-data)
	
	p = leastsq(err, p0, args=(data), maxfev=1000)
	p = p[0]
	
	#---------------------------------------------------------------------------------
	#formatting results
	maxi = data.max()
	floor = p[0]
	height = p[1]
	mean_x = p[2] + xy[0]
	mean_y = p[3] + xy[1]
	fwhm_small = np.min((np.abs(p[4]),np.abs(p[5])))
	fwhm_large = np.max((np.abs(p[4]),np.abs(p[5])))
	angle = p[6]	#gives the direction of the p[4]=sig_x axis, starting from x (vertical) axis, clockwise in direction of y (horizontal) axis
	if p[5]>p[4]:	#then sig_x is the smallest : we want angle to point to sig_y, the largest
		angle = angle + np.pi/2.
	angle = np.arctan(np.tan(angle))
	beta = p[7]
	
	output = [maxi, floor, height, mean_x, mean_y, fwhm_small, fwhm_large, angle, beta]
	return output
